from importlib import import_module, reload
from fire import Fire
from omegaconf import OmegaConf, DictConfig
from typing import Any
import sys
import os

# originally inspired by object instantiation code found in https://github.com/CompVis/taming-transformers. 
# added logic to recursively instantiate objects used as parameters, and recursively load omegaconfigs from yaml files to support superconfigs

SUPER_CONFIG_KEY = 'super'

OBJECT_PARAM_KEY = 'object_params'
CLASS_NAME_KEY = 'target' 
PARAM_KEY = 'params'
METHOD_KEY = 'method'

CONFIG_PATH_KEY = '_source_config_path'


# load omegaconfig from string filepath to yaml file
# if omegaconfig has SUPER_CONFIG_KEY key, recursively load omegaconfigs from corresponding value, which should be list of yaml filepaths to superconfigs
def load_config(yaml_file: str) -> DictConfig:

    config = OmegaConf.load(yaml_file)

    if SUPER_CONFIG_KEY in config:
        super_configs = []
        for super_config in config[SUPER_CONFIG_KEY]:
            super_configs.append(load_config(super_config))
        config = OmegaConf.unsafe_merge(*super_configs, config)
        
    return config


# instantiate object(s) specified by omegaconfig 
# if omegaconfig contains OBJECT_PARAM_KEY key, recursively instantiate objects from corresponding value, which should also be an omegaconfig
def instantiate_from_config(config: OmegaConf) -> Any:

    object_params = {}
    if OBJECT_PARAM_KEY in config:
        for object_name, object_config in config[OBJECT_PARAM_KEY].items():
            object_params[object_name] = instantiate_from_config(object_config)

    module_string, class_string = config[CLASS_NAME_KEY].rsplit(".", 1)
    if module_string in sys.modules:
        module = reload(sys.modules[module_string])
    else:
        module = import_module(module_string)
    module_class = getattr(module, class_string)

    # allow instantiation options like huggingface from_pretrained, instead of standard __init__
    # method should not use super(), see https://stackoverflow.com/questions/12047847/super-object-not-calling-getattr
    method_string = config.get(METHOD_KEY, '__init__') 
    if method_string is None:
        return module_class
    elif method_string == '__init__':
        return module_class(**config.get(PARAM_KEY, dict()), **object_params)
    method = getattr(module_class, method_string)
    return method(**config.get(PARAM_KEY, dict()), **object_params)

# instantiate object from yaml file, then run object method specified in method_string. 
# convenience function to run pytorch-lightning trainer methods from terminal
# i.e. to train, run "">> python lite.py path/to/omegaconfig.yaml fit"
def run(yaml_file: str, method_string: str=None, *args, **kwargs) -> Any:

    # add current directory to path for importing local modules
    sys.path.append(os.getcwd()) 

    config = load_config(yaml_file)
    object = instantiate_from_config(config)

    # attach source config path to object, avoid invoking unpredictable __setattr__ logic
    vars(object).update({CONFIG_PATH_KEY: yaml_file}) 
    
    if method_string is None:
        return object
    try:
        method = getattr(object, method_string)
        return object, method(*args, **kwargs)
    except KeyboardInterrupt:
        return object, None

def main():
    return Fire(run)

if __name__ == '__main__':
    main()



import __main__
from inspect import getmembers, ismethod, getfile
import torch
import os
from pathlib import Path
from pytorch_lightning.tuner.tuning import Tuner
from shutil import copy
from .lite import CONFIG_PATH_KEY
from pytorch_lightning.loggers.tensorboard import TensorBoardLogger

class Lab():
    def __init__(self, 
        trainer, 
        model=None, 
        datamodule=None, 
        ckpt_path=None, 
        state_dict_path=None, 
        **info
    ):
        vars(self).update(locals())
        self.tuner = Tuner(trainer)

        if ckpt_path == 'auto':
            ckpt_paths = self.get_ckpt_paths()
            self.ckpt_path = None
            if len(ckpt_paths) > 0:
                self.ckpt_path = ckpt_paths[-1]
            print(f'Checkpoint path set to {self.ckpt_path}')

        if state_dict_path is not None:
            self.load_state_dicts(state_dict_path)

        self.core = dict([(key, getattr(self, key)) for key in ['model', 'datamodule', 'ckpt_path']])

    def lr_find(self, *args, **kwargs):
        return self.tuner.lr_find(model=self.model, datamodule=self.datamodule, *args, **kwargs)

    def scale_batch_size(self, *args, **kwargs):
        return self.tuner.scale_batch_size(model=self.model, datamodule=self.datamodule, *args, **kwargs)

    # save a copy of relevant source files from project directory into log folder
    def save_source(self):
        log_source_dir = Path(self.log_dir)/'source'
        os.makedirs(log_source_dir, exist_ok=True)
        source_paths = [f'Checkpoint source: {self.ckpt_path}']

        if CONFIG_PATH_KEY in vars(self).keys():
            source_path = Path(vars(self)[CONFIG_PATH_KEY]).absolute()
            copy(source_path, log_source_dir)
            source_paths.append(f'Config source: {source_path}')

        for object in [self.model, self.datamodule, self]:
            source_path = getfile(object.__class__)
            copy(source_path, log_source_dir)
            source_paths.append(f'{object.__class__.__name__} source: {source_path}')

        with open(log_source_dir/'source_paths.txt', 'w+') as f:
            f.write('\n'.join(source_paths))
        return source_paths

    # assumes tensorboard logger directory structure. To do: check if this is consistent with other pl loggers
    def get_ckpt_paths(self, log_dir=None):
        if log_dir is None:
            log_dir = Path(self.log_dir).parent
        else:
            log_dir = Path(log_dir)
        ckpt_paths = list(log_dir.glob('checkpoints/*.ckpt')) + list(log_dir.glob('*/checkpoints/*.ckpt'))
        return ckpt_paths

    # restore checkpoints for testing/debugging in jupyter notebook
    def restore_checkpoint(self, ckpt_path=None):
        if ckpt_path is None:
            ckpt_path = self.ckpt_path
        self.trainer.state.fn = 'jupyter_notebook'
        self.trainer.strategy.connect(self.model)
        self.trainer._callback_connector._attach_model_callbacks()
        self.trainer._callback_connector._attach_model_logging_functions()
        self.trainer._checkpoint_connector.restore(ckpt_path)


    def load_state_dicts(self, file_paths=None, target_module=None, strict=False):

        if file_paths == None:
            ckpt_paths = self.get_ckpt_paths()
            if len(ckpt_paths) > 0:
                file_paths = [ckpt_paths[-1]]
            else:
                print('No state dicts loaded')
                return
        elif isinstance(file_paths, str):
            file_paths = [file_paths]

        if target_module is None:
            target_module = self.model

        for file_path in file_paths:
            state_dict = torch.load(file_path, map_location='cpu')
            if 'state_dict' in state_dict:
                state_dict = state_dict['state_dict']
            target_module.load_state_dict(state_dict, strict=strict)
            print(f'loaded state dict from {file_path} to {target_module.__class__.__name__}')


    def save_state_dict(self, save_path=None, as_checkpoint=False):
        if save_path is None:
            save_path = Path(self.log_dir)/f'{self.model.__class__.__name__}.pt'
        if as_checkpoint:
            return self.trainer.save_checkpoint(save_path)
        return torch.save(self.model.state_dict(), save_path)


    def __repr__(self):
        repr = super().__repr__() + '\n'
        for k, v in self.info.items():
            repr += f'{k}: {str(v)}\n'
        for k, v in self.core.items():
            repr += f'{k}: {v.__class__.__name__}\n'
        return repr


    def fit(self, *args, **kwargs):
        self.save_source()
        return self.trainer.fit(*args, **kwargs, **self.core)

    def validate(self, *args, **kwargs):
        self.save_source()
        return self.trainer.validate(*args, **kwargs, **self.core)

    def test(self, *args, **kwargs):
        self.save_source()
        return self.trainer.test(*args, **kwargs, **self.core)

    def predict(self, *args, **kwargs):
        self.save_source()
        return self.trainer.predict(*args, **kwargs, **self.core)

    # get batch from datamodule
    def get_batch(self, batch_idx=0):
        self.datamodule.prepare_data()
        try: self.datamodule.setup(stage='fit')
        except: pass
        batch_iters = iter(self.datamodule.train_dataloader())

        for _ in range(batch_idx):
            next(batch_iters)
        return next(batch_iters)

    # for easier attribute/variable access in jupyter notebook
    def unload(self, _to=__main__):
        vars(_to).update(vars(self))
        methods = getmembers(self, predicate=ismethod)

        for method_name, method in methods:
            setattr(_to, method_name, method)

    # reattach modified attributes/variables to self for further testing
    def reload(self, _from=__main__):
        vars(self).update({key: var for key, var in vars(_from).items() if key in vars(self)})
        methods = getmembers(_from, predicate=ismethod)

        for method_name, method in methods:
            if hasattr(self, method_name) and callable(getattr(self, method_name)):
                setattr(self, method_name, method)

    # same as trainer.log_dir but skip the strategy broadcast step
    # needed to access log_dir before trainer is intialized for multidevice strategy
    @property
    def log_dir(self):
        if len(self.trainer.loggers) > 0:
            if not isinstance(self.trainer.loggers[0], TensorBoardLogger):
                dirpath = self.trainer.loggers[0].save_dir
            else:
                dirpath = self.trainer.loggers[0].log_dir
        else:
            dirpath = self.trainer.default_root_dir
        return dirpath





from setuptools import setup, find_packages

repo_name = 'litelab'

# generate requirements file with !pipreqs litelab --mode no-pin --force
requirements_fn = "requirements.txt"
install_requirements = open(f"{repo_name}/{requirements_fn}").read().split()

setup(name=repo_name,
      version='0.0.3',
      description='Create lightweight configs for instantiating ML experiments',
      author='1lint',
      author_email='105617163+1lint@users.noreply.github.com',
      url=f'https://github.com/1lint/{repo_name}', 
      install_requires=install_requirements,
      packages=[repo_name],
      package_data={repo_name: [requirements_fn]},
      entry_points={
            'console_scripts': [
                  "lite = litelab:main"
            ]
      },
      )

